#include "Application.h"
#include "Benchmarker.h"
#include "ExampleFrameListener.h"
#include "InputManager.h"
#include "Logger.h"
#include "ParticleGroup.h"
#include "Renderer.h"
#include "Tools.h"

#include "OgreEntity.h"
#include "OgreLight.h"
#include "OgreMaterial.h"
#include "OgreMaterialManager.h"
#include "OgreOverlay.h"
#include "OgreOverlayElement.h"
#include "OgreOverlayManager.h"
#include "OgreRenderWindow.h"
#include "OgreRoot.h"
#include "OgreSceneManager.h"
#include "OgreSceneNode.h"
#include <sstream>


Application::Application():
	mCameraMode(CM_AUTOMATIC),
	mCameraAnimationPhase(0.0f),
	mCameraAnimationSpeed(2.0f),
	mInfoState(IS_FULL),
	mParticleRootNode(NULL),
	mParticleGroup(NULL),
	mParticleEmittersActive(true),
	currentMaterial(32),
	mVolumeDensity(5),
	mRenderer(NULL),
	mEnvironmentNode(NULL),
	mEnvironmentEntity(NULL),
	mLight0(NULL),
	mLight1(NULL),
	mBenchmarker(NULL)
{	
}


Application::~Application()
{
	delete mBenchmarker;

	if (mLight0)
	{
		mSceneMgr->destroyLight(mLight0);
	}
	if (mLight1)
	{
		mSceneMgr->destroyLight(mLight1);
	}

	if (mEnvironmentNode)
	{
		mEnvironmentNode->detachObject(mEnvironmentEntity);
		mSceneMgr->destroyEntity(mEnvironmentEntity);
		mEnvironmentNode->getParentSceneNode()
			->removeAndDestroyChild(mEnvironmentNode->getName());
	}

	delete mParticleGroup;
	if (mParticleRootNode)
	{
		mParticleRootNode->getParentSceneNode()
			->removeAndDestroyChild(mParticleRootNode->getName());
	}

	delete mRenderer;
	InputManager::get()->removeKeyListener(this);
}


void Application::createScene()
{
	//show the hud
	Ogre::Overlay* hud = Ogre::OverlayManager::getSingleton().getByName("HUD");
	hud->show();
	updateOverlayText();

	//set the skybox
	Ogre::Quaternion skyOrientation;
	skyOrientation.FromAxes(Ogre::Vector3::NEGATIVE_UNIT_X,
							Ogre::Vector3::UNIT_Y,
							Ogre::Vector3::NEGATIVE_UNIT_Z);
	mSceneMgr->setSkyBox(true, "Skybox", 75, true, skyOrientation);

	mParticleRootNode = mSceneMgr->getRootSceneNode()
		->createChildSceneNode("ParticleRootNode");

	//create the depth texture renderer
	mRenderer = new Renderer(mViewport, mWindow, mSceneMgr, mParticleRootNode, hud);

	//create all the materials that can be used
	materials.push_back(Material(mRenderer, "Planar unshaded particle without texture"									, "Plane.mesh"			, "UntexturedParticle"						, Material::RTM_NONE));
	materials.push_back(Material(mRenderer, "Planar mesh with 5 slices without texture"									, "Volume_5slices.mesh" , "Slices5"									, Material::RTM_NONE));
	materials.push_back(Material(mRenderer, "Planar mesh with 9 slices without texture"									, "Volume_9slices.mesh" , "Slices9"									, Material::RTM_NONE));
	materials.push_back(Material(mRenderer, "Planar mesh with 19 slices without texture"								, "Volume_19slices.mesh", "Slices19"								, Material::RTM_NONE));
	materials.push_back(Material(mRenderer, "Planar mesh with 39 slices without texture"								, "Volume_39slices.mesh", "Slices39"								, Material::RTM_NONE));
	materials.push_back(Material(mRenderer, "Planar mesh with 79 slices without texture"								, "Volume_79slices.mesh", "Slices79"								, Material::RTM_NONE));
	materials.push_back(Material(mRenderer, "Planar mesh with 5 slices without texture (depth texture still rendered)"	, "Volume_5slices.mesh" , "Slices5"									, Material::RTM_RENDER_WITHOUT_USAGE));
	materials.push_back(Material(mRenderer, "Planar mesh with 9 slices without texture (depth texture still rendered)"	, "Volume_9slices.mesh" , "Slices9"									, Material::RTM_RENDER_WITHOUT_USAGE));
	materials.push_back(Material(mRenderer, "Planar mesh with 19 slices without texture (depth texture still rendered)"	, "Volume_19slices.mesh", "Slices19"								, Material::RTM_RENDER_WITHOUT_USAGE));
	materials.push_back(Material(mRenderer, "Planar mesh with 39 slices without texture (depth texture still rendered)"	, "Volume_39slices.mesh", "Slices39"								, Material::RTM_RENDER_WITHOUT_USAGE));
	materials.push_back(Material(mRenderer, "Planar mesh with 79 slices without texture (depth texture still rendered)"	, "Volume_79slices.mesh", "Slices79"								, Material::RTM_RENDER_WITHOUT_USAGE));
	materials.push_back(Material(mRenderer, "Unshaded constant volume without texture"									, "Plane.mesh"			, "UnshadedVolumeConstantUntextured"		, Material::RTM_USAGE));
	materials.push_back(Material(mRenderer, "Unshaded lineair volume without texture"									, "Plane.mesh"			, "UnshadedVolumeLineairUntextured"			, Material::RTM_USAGE));
	materials.push_back(Material(mRenderer, "Unshaded spherical volume without texture, logarithmic density (default)"	, "Plane.mesh"			, "UnshadedVolumeSphereUntexturedLog"		, Material::RTM_USAGE));
	materials.push_back(Material(mRenderer, "Unshaded spherical volume without texture, lineair density"				, "Plane.mesh"			, "UnshadedVolumeSphereUntexturedLineair"	, Material::RTM_USAGE));
	materials.push_back(Material(mRenderer, "Unshaded spherical volume without texture, square root density"			, "Plane.mesh"			, "UnshadedVolumeSphereUntexturedSqrt"		, Material::RTM_USAGE));
	materials.push_back(Material(mRenderer, "Unshaded depth map volume without texture"									, "Plane.mesh"			, "UnshadedVolumeDepthmapUntextured"		, Material::RTM_USAGE));
	materials.push_back(Material(mRenderer, "Shaded spherical volume without texture"									, "Plane.mesh"			, "ShadedVolumeSphereUntextured"			, Material::RTM_USAGE));
	materials.push_back(Material(mRenderer, "Planar unshaded particle with texture"										, "Plane.mesh"			, "TexturedParticle"						, Material::RTM_NONE));
	materials.push_back(Material(mRenderer, "Planar mesh with 5 slices with texture"									, "Volume_5slices.mesh" , "Slices5Textured"							, Material::RTM_NONE));
	materials.push_back(Material(mRenderer, "Planar mesh with 9 slices with texture"									, "Volume_9slices.mesh" , "Slices9Textured"							, Material::RTM_NONE));
	materials.push_back(Material(mRenderer, "Planar mesh with 19 slices with texture"									, "Volume_19slices.mesh", "Slices19Textured"						, Material::RTM_NONE));
	materials.push_back(Material(mRenderer, "Planar mesh with 39 slices with texture"									, "Volume_39slices.mesh", "Slices39Textured"						, Material::RTM_NONE));
	materials.push_back(Material(mRenderer, "Planar mesh with 79 slices with texture"									, "Volume_79slices.mesh", "Slices79Textured"						, Material::RTM_NONE));
	materials.push_back(Material(mRenderer, "Planar mesh with 5 slices with texture (depth texture still rendered)"		, "Volume_5slices.mesh" , "Slices5Textured"							, Material::RTM_RENDER_WITHOUT_USAGE));
	materials.push_back(Material(mRenderer, "Planar mesh with 9 slices with texture (depth texture still rendered)"		, "Volume_9slices.mesh" , "Slices9Textured"							, Material::RTM_RENDER_WITHOUT_USAGE));
	materials.push_back(Material(mRenderer, "Planar mesh with 19 slices with texture (depth texture still rendered)"	, "Volume_19slices.mesh", "Slices19Textured"						, Material::RTM_RENDER_WITHOUT_USAGE));
	materials.push_back(Material(mRenderer, "Planar mesh with 39 slices with texture (depth texture still rendered)"	, "Volume_39slices.mesh", "Slices39Textured"						, Material::RTM_RENDER_WITHOUT_USAGE));
	materials.push_back(Material(mRenderer, "Planar mesh with 79 slices with texture (depth texture still rendered)"	, "Volume_79slices.mesh", "Slices79Textured"						, Material::RTM_RENDER_WITHOUT_USAGE));
	materials.push_back(Material(mRenderer, "Planar mesh with 5 slices with texture, rendered to floating buffer"		, "Volume_5slices.mesh" , "Slices5Textured"							, Material::RTM_RENDER_COLOUR_FLOAT));
	materials.push_back(Material(mRenderer, "Planar mesh with 9 slices with texture, rendered to floating buffer"		, "Volume_9slices.mesh" , "Slices9Textured"							, Material::RTM_RENDER_COLOUR_FLOAT));
	materials.push_back(Material(mRenderer, "Planar mesh with 19 slices with texture, rendered to floating buffer"		, "Volume_19slices.mesh", "Slices19Textured"						, Material::RTM_RENDER_COLOUR_FLOAT));
	materials.push_back(Material(mRenderer, "Planar mesh with 39 slices with texture, rendered to floating buffer"		, "Volume_39slices.mesh", "Slices39Textured"						, Material::RTM_RENDER_COLOUR_FLOAT));
	materials.push_back(Material(mRenderer, "Planar mesh with 79 slices with texture, rendered to floating buffer"		, "Volume_79slices.mesh", "Slices79Textured"						, Material::RTM_RENDER_COLOUR_FLOAT));
	materials.push_back(Material(mRenderer, "Unshaded constant volume with texture"										, "Plane.mesh"			, "UnshadedVolumeConstantTextured"			, Material::RTM_USAGE));
	materials.push_back(Material(mRenderer, "Unshaded lineair volume with texture"										, "Plane.mesh"			, "UnshadedVolumeLineairTextured"			, Material::RTM_USAGE));
	materials.push_back(Material(mRenderer, "Unshaded spherical volume with texture, logarithmic density (default)"		, "Plane.mesh"			, "UnshadedVolumeSphereTexturedLog"			, Material::RTM_USAGE));
	materials.push_back(Material(mRenderer, "Unshaded spherical volume with texture, lineair density"					, "Plane.mesh"			, "UnshadedVolumeSphereTexturedLineair"		, Material::RTM_USAGE));
	materials.push_back(Material(mRenderer, "Unshaded spherical volume with texture, square root density"				, "Plane.mesh"			, "UnshadedVolumeSphereTexturedSqrt"		, Material::RTM_USAGE));
	materials.push_back(Material(mRenderer, "Unshaded depth map volume with texture"									, "Plane.mesh"			, "UnshadedVolumeDepthmapTextured"			, Material::RTM_USAGE));
	materials.push_back(Material(mRenderer, "Shaded spherical volume with texture"										, "Plane.mesh"			, "ShadedVolumeSphereTextured"				, Material::RTM_USAGE));

	//create particle system
	mParticleGroup = new ParticleGroup(mParticleRootNode,
		materials[currentMaterial].mesh, materials[currentMaterial].material,
		materials[currentMaterial].renderTextureMode == Material::RTM_USAGE,
		mParticleEmittersActive, mVolumeDensity, mRenderer);

	//create a scene for the effects to happen in
	mEnvironmentEntity = mSceneMgr
		->createEntity("Environment" + Tools::getUniqueString(), "BoxGrid.mesh");
	mEnvironmentNode = mSceneMgr->getRootSceneNode()
		->createChildSceneNode("Environment" + Tools::getUniqueString());
	mEnvironmentNode->attachObject(mEnvironmentEntity);

	//create two lights to light the scene
	mLight0 = mSceneMgr->createLight("Light0");
	mLight0->setPosition(Ogre::Vector3(30, 20, 10));
	mLight0->setDiffuseColour(Ogre::ColourValue(1, 1, 1));
	mLight1 = mSceneMgr->createLight("Light");
	mLight1->setPosition(Ogre::Vector3(-20, -10, -30));
	mLight1->setDiffuseColour(Ogre::ColourValue(0.7f, 0.7f, 0.7f));

	//start listening to input events
	InputManager::get()->addKeyListener(this);
}


void Application::createFrameListener()
{
	mFrameListener = new ExampleFrameListener(mWindow, mCamera);
	mRoot->addFrameListener(this);
}


void Application::keyPressed(const OIS::KeyEvent& e)
{
	int key = e.key;
	switch (key)
	{
	case OIS::KC_F1:
		mInfoState = InfoState((mInfoState + 1) % 3);
		break;
	case OIS::KC_COMMA:
		setCurrentMaterialIndex((currentMaterial + getNumMaterials() - 1) % getNumMaterials());
		break;
	case OIS::KC_PERIOD:
		setCurrentMaterialIndex((currentMaterial + 1) % getNumMaterials());
		break;
	case OIS::KC_P:
		switchParticleEmittersActive();
		break;
	case OIS::KC_F5:
		if (!mBenchmarker)
		{
			mBenchmarker = new Benchmarker(this, mCamera);
		}
		break;
	}

	//handle swapping camera modes
	if (mCameraMode == CM_FREE)
	{
		if (e.key == OIS::KC_C)
		{
			mCameraMode = CM_AUTOMATIC;
			mFrameListener->setCameraRotationAllowed(false);
		}
	}
	else
	{
		//switch to the free camera if the user tries to control the camera
		if (e.key == OIS::KC_C
			|| e.key == OIS::KC_W
			|| e.key == OIS::KC_A
			|| e.key == OIS::KC_S
			|| e.key == OIS::KC_D
			|| e.key == OIS::KC_PGDOWN
			|| e.key == OIS::KC_PGUP
			|| e.key == OIS::KC_LEFT
			|| e.key == OIS::KC_RIGHT
			|| e.key == OIS::KC_UP
			|| e.key == OIS::KC_DOWN)
		{
			mCameraMode = CM_FREE;
			mFrameListener->setCameraRotationAllowed(true);
		}
	}
}


bool Application::frameStarted(const Ogre::FrameEvent& evt)
{
	//update user input through keyboard and mouse
	bool keepRunning = mFrameListener->frameStarted(evt);

	float time = evt.timeSinceLastFrame;

	//check whether benchmarking is happening right now
	if (mBenchmarker)
	{
		//busy benchmarking, so update that process
		mBenchmarker->update(time);
		if (mBenchmarker->isFinished())
		{
			//end benchmarking
			delete mBenchmarker;
			mBenchmarker = NULL;
		}
	}
	else
	{
		//perform a standard update
		updateCameraAnimation(time);
		updateKeysDown(time);
		updateOverlayText();
	}
	mParticleGroup->update(time, mCamera->getPosition());

	return keepRunning && !mWindow->isClosed();
}


bool Application::frameEnded(const Ogre::FrameEvent& evt)
{
	return mFrameListener->frameEnded(evt);
}


void Application::setOverlayTexts(const std::string& text0,
								  const std::string& text1) const
{
	Ogre::OverlayManager& overlayManager = Ogre::OverlayManager::getSingleton();
	overlayManager.getOverlayElement("HUD/Materials")->setCaption(text0);
	overlayManager.getOverlayElement("HUD/Buttons")->setCaption(text1);
}


void Application::setParticleSprayActive(bool active)
{
	mParticleGroup->setEmittersActive(active);
}


unsigned int Application::getNumMaterials() const
{
	return (unsigned int)(materials.size());
}


void Application::setCurrentMaterialIndex(unsigned int index)
{
	if (index < materials.size()
		&& index != currentMaterial)
	{
		currentMaterial = index;

		//perform the actual swapping of the material of the particles
		bool useRenderTexture =
			materials[currentMaterial].renderTextureMode == Material::RTM_USAGE;
		mParticleGroup->setVolumeMode(materials[currentMaterial].mesh,
									materials[currentMaterial].material,
									useRenderTexture);

		switch (materials[currentMaterial].renderTextureMode)
		{
		case Material::RTM_USAGE:
			mRenderer->setActive(true);
			mRenderer->setRenderColourFloatOverlay(false);
			break;
		case Material::RTM_RENDER_WITHOUT_USAGE:
			mRenderer->setActive(true);
			mRenderer->setRenderColourFloatOverlay(false);
			break;
		case Material::RTM_RENDER_COLOUR_FLOAT:
			mRenderer->setActive(true);
			mRenderer->setRenderColourFloatOverlay(true);
			break;
		case Material::RTM_NONE:
			mRenderer->setActive(false);
			mRenderer->setRenderColourFloatOverlay(false);
			break;
		}
	}
}


unsigned int Application::getCurrentMaterialIndex() const
{
	return currentMaterial;
}


std::string Application::getCurrentMaterialName() const
{
	return materials[currentMaterial].description;
}


int Application::getVersionNumber() const
{
	return 7;
}


void Application::switchParticleEmittersActive()
{
	mParticleEmittersActive = !mParticleEmittersActive;
	mParticleGroup->setEmittersActive(mParticleEmittersActive);
}


void Application::setVolumeDensity(float density)
{
	mVolumeDensity = density;
	mParticleGroup->setVolumeDensity(mVolumeDensity);
}


void Application::updateKeysDown(float time)
{
	InputManager* inputManager = InputManager::get();

	//handle changing the density of the volumes
	if (inputManager->isKeyDown(OIS::KC_LBRACKET)) setVolumeDensity(mVolumeDensity * pow(0.5f, time));
	if (inputManager->isKeyDown(OIS::KC_RBRACKET)) setVolumeDensity(mVolumeDensity * pow(2.0f, time));
}


void Application::updateCameraAnimation(float time)
{
	if (mCameraMode == CM_AUTOMATIC)
	{
		mCameraAnimationPhase += time * mCameraAnimationSpeed;
		Ogre::Vector3 localPosition(sin(0.25f * mCameraAnimationPhase),
									cos(0.11f * mCameraAnimationPhase) * 0.25f,
									cos(0.25f * mCameraAnimationPhase));
		Ogre::Vector3 centre(0, 1, 0);
		mCamera->setPosition(10 * localPosition + centre);
		mCamera->lookAt(centre);
	}
}


void Application::updateOverlayText() const
{
	std::stringstream text0, text1;
	if (mInfoState != IS_NONE)
	{
		text0 << "Volumetrics version " << getVersionNumber();
	}
	if (mInfoState == IS_FULL)
	{
		updateFullInformation(text0, text1);
	}
	else if (mInfoState == IS_SMALL)
	{
		updateSmallInformation(text0);
	}
	setOverlayTexts(text0.str(), text1.str());
}


void Application::updateFullInformation(std::stringstream& text0,
										std::stringstream& text1) const
{
	//write a list of all VolumeModes
	text0 << "\n\nVolume modes:";
	for (unsigned int i = 0; i < materials.size(); ++i)
	{
		std::string description = materials[i].description;
		if (i == currentMaterial)
		{
			text0 << "\n >" << Tools::toUpperCase(description);
		}
		else
		{
			text0 << "\n  " << description;
		}
	}

	text1
		<< "Keys:"
		<< "\n  F1: hide information"
		<< "\n  F: show/hide framerate"
		<< "\n  , (comma) and . (period): switch volume type"
		<< "\n  [ and ]: change volume density"
		<< "\n  P: switch particle emitting"
		<< "\n  R: wireframe"
		<< "\n  Print Screen: save screenshot"
		<< "\n  WASD: move camera"
		<< "\n  F5: Start benchmark";

	text1
		<< "\n  C: "
		<< (mCameraMode == CM_FREE ? "switch to automatic camera" : "switch to free camera");

	text1
		<< "\n\nParticle emitters: "
		<< (mParticleEmittersActive ? "on" : "off");

	text1
		<< "\nVolume density: "
		<< mVolumeDensity;
}


void Application::updateSmallInformation(std::stringstream& text0) const
{
	text0
		<< "\nCurrent volume mode: "
		<< materials[currentMaterial].description
		<< "\nPress 'F1' for more information.";
}