#include "InputManager.h"
#include "KeyListener.h"
#include "Logger.h"
#include "MouseListener.h"

#include "OISInputManager.h"
#include "OgreRenderWindow.h"
#include "OgreRoot.h"


InputManager* InputManager::get()
{
	static InputManager* instance = NULL;
	if (!instance)
	{
		instance = new InputManager();
	}
	return instance;
}


void InputManager::capture()
{
	keyboard->capture();
	mouse->capture();
}


bool InputManager::isKeyDown(OIS::KeyCode keyCode) const
{
	if (exclusiveKeyListener)
	{
		return false;
	}
	else
	{
		return keyboard->isKeyDown(keyCode);
	}
}


bool InputManager::isMouseButtonDown(MouseButton mouseButton) const
{
	if (exclusiveMouseListener)
	{
		return false;
	}
	else
	{
		switch (mouseButton)
		{
			case MB_LEFT:
				return mouse->getMouseState().buttonDown(OIS::MB_Left);
				break;
			case MB_MIDDLE:
				return mouse->getMouseState().buttonDown(OIS::MB_Middle);
				break;
			case MB_RIGHT:
				return mouse->getMouseState().buttonDown(OIS::MB_Right);
				break;
		}
		return false;
	}
}


int InputManager::getMouseRelativeX()
{
	if (exclusiveMouseListener)
	{
		return 0;
	}
	else
	{
		return mouse->getMouseState().X.rel;
	}
}


int InputManager::getMouseRelativeY()
{
	if (exclusiveMouseListener)
	{
		return 0;
	}
	else
	{
		return mouse->getMouseState().Y.rel;
	}
}


void InputManager::addKeyListener(::KeyListener* keyListener)
{
	keyListeners.insert(keyListener);
}


void InputManager::removeKeyListener(::KeyListener* keyListener)
{
	keyListeners.erase(keyListener);
}


void InputManager::clearAllKeyListeners()
{
	keyListeners.clear();
}


void InputManager::setExclusiveKeyListener(::KeyListener* keyListener)
{
	exclusiveKeyListener = keyListener;
}


void InputManager::removeExclusiveKeyListener()
{
	exclusiveKeyListener = NULL;
}


void InputManager::addMouseListener(::MouseListener* mouseListener)
{
	mouseListeners.insert(mouseListener);
}


void InputManager::removeMouseListener(::MouseListener* mouseListener)
{
	mouseListeners.erase(mouseListener);
}


void InputManager::clearAllMouseListeners()
{
	mouseListeners.clear();
}


void InputManager::setExclusiveMouseListener(::MouseListener* mouseListener)
{
	exclusiveMouseListener = mouseListener;
}


void InputManager::removeExclusiveMouseListener()
{
	exclusiveMouseListener = NULL;
}


bool InputManager::keyPressed(const OIS::KeyEvent& e)
{
	if (exclusiveKeyListener)
	{
		exclusiveKeyListener->keyPressed(e);
	}
	else
	{
		std::set<::KeyListener*>::iterator iter, end;
		iter = keyListeners.begin();
		end = keyListeners.end();
		while (iter != end)
		{
			(*iter)->keyPressed(e);
			++iter;
		}
	}
	return true;
}


bool InputManager::keyReleased(const OIS::KeyEvent& e)
{
	if (exclusiveKeyListener)
	{
		exclusiveKeyListener->keyReleased(e);
	}
	else
	{
		std::set<::KeyListener*>::iterator iter, end;
		iter = keyListeners.begin();
		end = keyListeners.end();
		while (iter != end)
		{
			(*iter)->keyReleased(e);
			++iter;
		}
	}
	return true;
}


bool InputManager::mouseMoved(const OIS::MouseEvent& e)
{
	if (exclusiveMouseListener)
	{
		exclusiveMouseListener->mouseMoved(e);
	}
	else
	{
		std::set<::MouseListener*>::iterator iter, end;
		iter = mouseListeners.begin();
		end = mouseListeners.end();
		while (iter != end)
		{
			(*iter)->mouseMoved(e);
			++iter;
		}
	}
	return true;
}


bool InputManager::mousePressed(const OIS::MouseEvent& e, OIS::MouseButtonID id)
{
	if (exclusiveMouseListener)
	{
		exclusiveMouseListener->mousePressed(e, id);
	}
	else
	{
		std::set<::MouseListener*>::iterator iter, end;
		iter = mouseListeners.begin();
		end = mouseListeners.end();
		while (iter != end)
		{
			(*iter)->mousePressed(e, id);
			++iter;
		}
	}
	return true;
}


bool InputManager::mouseReleased(const OIS::MouseEvent& e, OIS::MouseButtonID id)
{
	if (exclusiveMouseListener)
	{
		exclusiveMouseListener->mouseReleased(e, id);
	}
	else
	{
		std::set<::MouseListener*>::iterator iter, end;
		iter = mouseListeners.begin();
		end = mouseListeners.end();
		while (iter != end)
		{
			(*iter)->mouseReleased(e, id);
			++iter;
		}
	}
	return true;
}


InputManager::InputManager():
	inputManager(NULL),
	keyboard(NULL),
	mouse(NULL),
	exclusiveKeyListener(NULL),
	exclusiveMouseListener(NULL)
{
	size_t hWnd = 0;
	Ogre::RenderWindow* window = Ogre::Root::getSingleton().getAutoCreatedWindow();
	if (window)
	{
		window->getCustomAttribute("WINDOW", &hWnd);
		inputManager = OIS::InputManager::createInputSystem(hWnd);

		keyboard = static_cast<OIS::Keyboard*>(inputManager->createInputObject(OIS::OISKeyboard, true));
		keyboard->setTextTranslation(OIS::Keyboard::Ascii);
		keyboard->setEventCallback(this);

		mouse = static_cast<OIS::Mouse*>(inputManager->createInputObject(OIS::OISMouse, true));
		mouse->setEventCallback(this);
	}
}


InputManager::~InputManager()
{
	delete keyboard;
	delete mouse;
	if (inputManager)
	{
		OIS::InputManager::destroyInputSystem(inputManager);
	}
}