#include "MeshParticle.h"
#include "Renderer.h"
#include "Tools.h"

#include "OgreEntity.h"
#include "OgreMaterial.h"
#include "OgreMaterialManager.h"
#include "OgreSceneManager.h"
#include "OgreSceneNode.h"
#include "OgreSubEntity.h"


MeshParticle::MeshParticle(Ogre::SceneNode* root,
						   const Ogre::Vector3& position,
						   const Ogre::Vector3& speed,
						   float lifeTime,
						   float scale,
						   const std::string& mesh,
						   const std::string& materialName,
						   bool materialRequiresDepthTexture,
						   const std::string& textureName,
						   const std::string& thicknessTextureName,
						   float volumeDensity,
						   bool animateParameters,
						   Renderer* renderer):
	mRenderer(renderer),
	mSpeed(speed),
	mLifeTime(lifeTime),
	mTimeExisting(0),
	mInitialScale(scale),
	mCurrentScale(mInitialScale),
	mMesh(mesh),
	mMaterialName(materialName),
	mMaterialRequiresDepthTexture(materialRequiresDepthTexture),
	mTextureName(textureName),
	mThicknessTextureName(thicknessTextureName),
	mVolumeDensity(volumeDensity),
	mAnimateParameters(animateParameters)
{
	mID = "P" + Tools::getUniqueString();

	mNode = root->createChildSceneNode(mID);
	mNode->setPosition(position);

	loadVisualisation();
}


MeshParticle::~MeshParticle()
{
	removeVisualisation();
	mNode->getParentSceneNode()->removeAndDestroyChild(mNode->getName());
}


void MeshParticle::update(float time, const Ogre::Vector3& viewPosition)
{
	mTimeExisting += time;
	mNode->translate(mSpeed * time, Ogre::Node::TS_WORLD);

	//Scale is multiplied to make the particle grow during its lifetime
	if (mAnimateParameters)
	{
		float timeFraction = mTimeExisting / mLifeTime;
		float fadeInFraction = 0.5f;
		if (timeFraction < fadeInFraction)
		{
			//smoothly grow the particle during the first phase of its existince
			mCurrentScale = cos(timeFraction / fadeInFraction * 3.1415f) * -0.5f + 0.5f;
		}
		else
		{
			//slowly grow during the rest of its existence
			mCurrentScale = 1 + (timeFraction - fadeInFraction) / (1 - fadeInFraction);
		}
		mCurrentScale *= mInitialScale;
	}

	mNode->setScale(Ogre::Vector3(mCurrentScale));

	//orient particle towards camera
	Ogre::Vector3 dir = viewPosition - mNode->getPosition();
	dir.normalise();
	Ogre::Vector3 side = dir.crossProduct(Ogre::Vector3(0, 1, 0));
	side.normalise();
	Ogre::Vector3 up = side.crossProduct(dir);
	mNode->setOrientation(Ogre::Quaternion(-up, -dir, -side));

	setEntityParameters();
}


bool MeshParticle::wantsToBeDestroyed() const
{
	return mTimeExisting >= mLifeTime;
}


void MeshParticle::setVolumeMode(const std::string& mesh,
								 const std::string& material,
								 bool materialRequiresDepthTexture)
{
	mMesh = mesh;
	mMaterialName = material;
	mMaterialRequiresDepthTexture = materialRequiresDepthTexture;
	removeVisualisation();
	loadVisualisation();
}


void MeshParticle::setVolumeDensity(float density)
{
	mVolumeDensity = density;
}


void MeshParticle::removeVisualisation()
{
	if (mEntity)
	{
		mNode->detachObject(mEntity);
		mNode->getCreator()->destroyEntity(mEntity);
		mEntity = NULL;

		Ogre::MaterialManager::getSingleton().remove(mMaterial->getName());
		mMaterial = NULL;
	}
}


void MeshParticle::loadVisualisation()
{
	//create the entity
	mEntity = mNode->getCreator()->createEntity(mID, mMesh);
	mNode->attachObject(mEntity);

	Ogre::ResourcePtr resource =
		Ogre::MaterialManager::getSingleton().getByName(mMaterialName);

	mMaterial =	static_cast<Ogre::MaterialPtr>(resource)
		->clone(mMaterialName + "_" + mID).getPointer();

	//set the correct texture in the material
	Ogre::Pass* pass = mMaterial->getTechnique(0)->getPass(0);
	for (size_t i = 0; i < pass->getNumTextureUnitStates(); ++i)
	{
		Ogre::TextureUnitState* texture = pass->getTextureUnitState((unsigned short)(i));

		if (texture->getTextureName() == "Smoke04.tga")
		{
			texture->setTextureName(mTextureName);
		}
		if (texture->getTextureName() == "Smoke04_thickness.tga")
		{
			texture->setTextureName(mThicknessTextureName);
		}
	}

	if (mMaterialRequiresDepthTexture)
	{
		mRenderer->addMaterialWithDepth(mMaterial->getName());
	}

	//apply the newly created material to the object
	mEntity->setMaterialName(mMaterial->getName());
}


void MeshParticle::setEntityParameters()
{
	if (mEntity)
	{
		//calculate what percentage of the lifetime of this particles has passed
		float lifeTimeFraction = mTimeExisting / mLifeTime;

		Ogre::Vector3 position = mNode->getWorldPosition();

		//Scale is divided by 2, because it is a radius, and then divided by a little more
		//because it must fit inside the plane.
		float scale = mCurrentScale / 2.2f;

		//Pack position and scale into a parameter that is to be passed to the pixel shader.
		Ogre::Vector4 position_scale(position.x, position.y, position.z, scale);

		Ogre::Vector4 density(1, 0, 0, 0);

		if (mAnimateParameters)
		{
			//make the density decrease during the lifetime of the particle
			density[0] = 1 - (lifeTimeFraction - 0.5f) * 2;
			density[0] = std::min(1.0f, density[0]);
		}		
		density[0] *= mVolumeDensity;

		//set the parameters on all SubEntities
		unsigned int numSubEntities = mEntity->getNumSubEntities();
		for (unsigned int i = 0; i < numSubEntities; ++i)
		{
			mEntity->getSubEntity(i)->setCustomParameter(0, position_scale);
			mEntity->getSubEntity(i)->setCustomParameter(1, density);
		}
	}
}