#include "ParticleGroup.h"
#include "MeshParticle.h"
#include "Tools.h"

#include "OgreSceneManager.h"
#include "OgreSceneNode.h"


std::vector<std::pair<std::string, std::string> > ParticleGroup::sTextures;


ParticleGroup::ParticleGroup(Ogre::SceneNode* particleRootNode,
							 const std::string& mesh,
							 const std::string& material,
							 bool materialRequiresDepthTexture,
							 bool emittersActive,
							 float volumeDensity,
							 Renderer* renderer):
	mMesh(mesh),
	mMaterial(material),
	mMaterialRequiresDepthTexture(materialRequiresDepthTexture),
	mEmittersActive(emittersActive),
	mTimeUntilNextEmitting(0),
	mTimeBetweenParticleEmitting(0.3f),
	mVolumeDensity(volumeDensity),
	mRenderer(renderer)
{
	//add all textures that can be used, if the static list is still empty
	if (sTextures.size() == 0)
	{
		//sTextures.push_back(std::make_pair("Smoke03.tga", "Smoke03_thickness.tga"));
		sTextures.push_back(std::make_pair("Smoke08.tga", "Smoke08_thickness.tga"));
	}

	//create a root SceneNode for all the particles in this group to hang on
	mGroupRoot = particleRootNode
		->createChildSceneNode("ParticleGroupRoot" + Tools::getUniqueString());

	//create a particle to constantly live at the centre of the scene
	createParticle(Ogre::Vector3(0, 0, 0), Ogre::Vector3(0, 0, 0), 1e10, 6, false);
}


ParticleGroup::~ParticleGroup()
{
	size_t numParticles = mParticles.size();
	for (size_t i = 0; i < numParticles; ++i)
	{
		delete mParticles[i];
	}

	mGroupRoot->getParentSceneNode()->removeAndDestroyChild(mGroupRoot->getName());
}


void ParticleGroup::update(float time, const Ogre::Vector3& viewPosition)
{
	//handle emitting of new particles
	updateParticleEmitting(time);

	//update all existing particles
	int numParticles = (int)(mParticles.size());
	for (int i = 0; i < numParticles; ++i)
	{
		//update the particle
		mParticles[i]->update(time, viewPosition);

		//check whether the particle should be destroyed
		if (mParticles[i]->wantsToBeDestroyed())
		{
			//delete the particle
			delete mParticles[i];

			//remove the particle from the list
			mParticles.erase(mParticles.begin() + i);

			//update iterator to fit changed list
			i--;
			numParticles--;
		}
	}
}


void ParticleGroup::setVolumeMode(const std::string& mesh,
								  const std::string& material,
								  bool materialRequiresDepthTexture)
{
	mMesh = mesh;
	mMaterial = material;
	mMaterialRequiresDepthTexture = materialRequiresDepthTexture;

	size_t numParticles = mParticles.size();
	for (size_t i = 0; i < numParticles; ++i)
	{
		mParticles[i]->setVolumeMode(mMesh, mMaterial, mMaterialRequiresDepthTexture);
	}
}


void ParticleGroup::setEmittersActive(bool active)
{
	mEmittersActive = active;
	if (!active)
	{
		//destroy all particles except the first one
		for (size_t i = 1; i < mParticles.size(); ++i)
		{
			delete mParticles[i];
			mParticles.erase(mParticles.begin() + i);
			i--;
		}
	}
}


void ParticleGroup::setVolumeDensity(float density)
{
	mVolumeDensity = density;
	size_t numParticles = mParticles.size();
	for (size_t i = 0; i < numParticles; ++i)
	{
		mParticles[i]->setVolumeDensity(mVolumeDensity);
	}
}


void ParticleGroup::updateParticleEmitting(float time)
{
	mTimeUntilNextEmitting -= time;

	if (mEmittersActive && mTimeUntilNextEmitting <= 0)
	{
		//calculate a random direction and a random speed
		Ogre::Vector3 speed;
		speed.x = Tools::getRandom(-1, 1);
		speed.y = Tools::getRandom(2, 6);
		speed.z = Tools::getRandom(-1, 1);
		speed.normalise();
		speed *= Tools::getRandom(0.5f, 0.6f);

		//calculate a random lifeTime and scale
		float lifeTime = Tools::getRandom(5, 20);
		float scale = Tools::getRandom(1, 2);

		//create the actual particle
		createParticle(Ogre::Vector3(0, 3, 0), speed, lifeTime, scale, true);

		//wait a bit before the next particle is to be created
		mTimeUntilNextEmitting = mTimeBetweenParticleEmitting;
	}
}


void ParticleGroup::createParticle(const Ogre::Vector3& position,
								   const Ogre::Vector3& speed,
								   float lifeTime,
								   float scale,
								   bool animateParemeters)
{
	size_t index = std::rand() % sTextures.size();

	mParticles.push_back(new MeshParticle(mGroupRoot, position,
		speed, lifeTime, scale, mMesh, mMaterial,
		mMaterialRequiresDepthTexture,
		sTextures[index].first, sTextures[index].second,
		mVolumeDensity, animateParemeters, mRenderer));
}