#include "Renderer.h"
#include "Logger.h"
#include "Tools.h"

#include "OgreEntity.h"
#include "OgreHardwarePixelBuffer.h"
#include "OgreMaterial.h"
#include "OgreMaterialManager.h"
#include "OgreOverlayManager.h"
#include "OgreRenderTexture.h"
#include "OgreRenderWindow.h"
#include "OgreRoot.h"
#include "OgreSceneNode.h"
#include "OgreSubEntity.h"
#include <vector>


Renderer::Renderer(Ogre::Viewport* screenViewport,
				   Ogre::RenderWindow* renderWindow,
				   Ogre::SceneManager* sceneManager,
				   Ogre::SceneNode* particleRootNode,
				   Ogre::Overlay* hudOverlay):
	mScreenViewport(screenViewport),
	mRenderWindow(renderWindow),
	mSceneManager(sceneManager),
	mParticleRootNode(particleRootNode),
	mHudOverlay(hudOverlay),
	mIsActive(true),
	mRenderColourFloatOverlay(false)
{
	//fetch stuff needed for intializing post effect
	Ogre::OverlayManager& overlayManager = Ogre::OverlayManager::getSingleton();
	Ogre::RenderSystem* renderSystem = Ogre::Root::getSingleton().getRenderSystem();

	//get render-texture resolution
	int width = Tools::round(renderWindow->getWidth() * screenViewport->getWidth());
	int height = Tools::round(renderWindow->getHeight() * screenViewport->getHeight());

	//create the RenderTextures
	mDepthTexture = *createRenderTexture(width, height, Ogre::PF_FLOAT16_R);
	mColourFloatTexture = *createRenderTexture(width, height, Ogre::PF_FLOAT16_RGB);

	//add the colour texture to the material
	static_cast<Ogre::MaterialPtr>(Ogre::MaterialManager::getSingleton().getByName("ColourFloatMaterial"))
		->getTechnique(0)->getPass(0)
		->createTextureUnitState(mColourFloatTexture.texture->getName());

	//get the debugging overlay, as it needs to be hidden and shown at the right moments
	mDebugOverlay = overlayManager.getByName("Core/DebugOverlay");
	mColourFloatOverlay = overlayManager.getByName("ColourFloatOverlay");

	//listen for rendering events
	mRenderWindow->addListener(this);
}


Renderer::~Renderer()
{
	//stop listening for rendering events
	mRenderWindow->removeListener(this);

	destroyRenderTexture(mDepthTexture);
	destroyRenderTexture(mColourFloatTexture);
}


void Renderer::preViewportUpdate(const Ogre::RenderTargetViewportEvent& evt)
{
	if (evt.source == mScreenViewport && mIsActive)
	{
		if (mRenderColourFloatOverlay)
		{
			handleSceneToFloatBufferRendering();
		}
		else
		{
			handleDepthTextureRendering();
		}
	}
}


void Renderer::postViewportUpdate(const Ogre::RenderTargetViewportEvent& evt)
{
	if (evt.source == mScreenViewport && mIsActive)
	{
		mColourFloatOverlay->hide();
		
		//reattach any objects that were detached just before rendering
		reattachAllObjects();
	}
}


void Renderer::addMaterialWithDepth(const std::string& material)
{
	addTextureToMaterial(mDepthTexture, material);
}


void Renderer::setActive(bool active)
{
	mIsActive = active;
}


void Renderer::setRenderColourFloatOverlay(bool renderColourFloatOverlay)
{
	mRenderColourFloatOverlay = renderColourFloatOverlay;
}


Renderer::RenderTexture* Renderer::createRenderTexture(unsigned int width,
													   unsigned int height,
													   Ogre::PixelFormat format)
{
	RenderTexture* result = new RenderTexture();

	Ogre::TexturePtr texture = Ogre::TextureManager::getSingleton().createManual(
		Tools::append("RenderTexture", Tools::getUniqueNumber()),
		Ogre::ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME,
		Ogre::TEX_TYPE_2D, 
		width,
		height,
		0,
		format,
		Ogre::TU_RENDERTARGET,
		NULL);
	result->texture = texture.getPointer();
	result->renderTarget = texture->getBuffer()->getRenderTarget();
	result->renderTarget->setAutoUpdated(false);
	result->viewport = result->renderTarget->addViewport(mScreenViewport->getCamera());
	result->viewport->setClearEveryFrame(true);
	result->viewport->setBackgroundColour(Ogre::ColourValue::Black);

	return result;
}


void Renderer::destroyRenderTexture(RenderTexture& renderTexture)
{
	renderTexture.renderTarget->removeViewport(renderTexture.viewport->getZOrder());
	Ogre::TextureManager::getSingleton().remove(renderTexture.texture->getName());
}


void Renderer::addTextureToMaterial(const RenderTexture& renderTexture,
									const std::string materialName)
{
	//get the material
	Ogre::MaterialPtr material =
		static_cast<Ogre::MaterialPtr>(Ogre::MaterialManager::getSingleton().getByName(materialName));

	//check whether the material exists
	if (!material.isNull())
	{
		//get the render pass
		Ogre::Pass* pass = material->getTechnique(0)->getPass(0);

		//add the RenderTexture to the render pass
		Ogre::TextureUnitState* texture = 		
			pass->createTextureUnitState(renderTexture.texture->getName());

		//the RenderTexture should not be filtered or mipmapped
		texture->setTextureFiltering(Ogre::TFO_NONE);

		//store the size of a half pixel in the pixel shader to correct
		size_t width = renderTexture.texture->getWidth();
		size_t height = renderTexture.texture->getHeight();
		pass->getFragmentProgramParameters()
			->setNamedConstant("uvOffset", Ogre::Vector3(0.5f / width, 0.5f / height, 0));
	}
	else
	{
		//the material does not exist, so print an error and do nothing else
		Logger::log("ERROR: Renderer::addTextureToMaterial: cannot find material",
					materialName);
	}
}


void Renderer::handleDepthTextureRendering()
{
	//swap materials of all objects to depth materials and store their original materials
	std::vector<std::pair<Ogre::SubEntity*, std::string> > originalMaterials;
	//walk over all Entities
	Ogre::SceneManager::MovableObjectIterator iter = mSceneManager->getMovableObjectIterator("Entity");
	while (iter.hasMoreElements())
	{
		Ogre::Entity* entity = static_cast<Ogre::Entity*>(iter.getNext());

		//walk over all SubEntities of this Entity
		unsigned int numSubEntities = entity->getNumSubEntities();
		for (unsigned int i = 0; i < numSubEntities; ++i)
		{
			Ogre::SubEntity* subEntity = entity->getSubEntity(i);

			//store the original material of this subEntity
			originalMaterials.push_back(std::make_pair(subEntity, subEntity->getMaterialName()));

			//swap this SubEntity's material to the depth rendering material
			subEntity->setMaterialName("Depth");
		}
	}

	//store whether overlays were visible before to be able to
	//correctly restore the status afterwards
	bool debugOverlayVisible = mDebugOverlay->isVisible();
	bool hudOverlayVisible = mHudOverlay->isVisible();

	//hide overlays
	mDebugOverlay->hide();
	mHudOverlay->hide();

	//hide the particles that will use the depth RenderTexture to render
	mParticleRootNode->setVisible(false);

	//render the depth to the RenderTexture
	mDepthTexture.renderTarget->update();

	//show the particles that will use the depth RenderTexture to render
	mParticleRootNode->setVisible(true);

	//show overlays that were visible
	if (debugOverlayVisible)
	{
		mDebugOverlay->show();
	}
	if (hudOverlayVisible)
	{
		mHudOverlay->show();
	}

	//swap all materials back to their original states
	size_t numSubEntities = originalMaterials.size();
	for (size_t i = 0; i < numSubEntities; ++i)
	{
		originalMaterials[i].first->setMaterialName(originalMaterials[i].second);
	}
}


void Renderer::handleSceneToFloatBufferRendering()
{
	//store whether overlays were visible before to be able to
	//correctly restore the status afterwards
	bool debugOverlayVisible = mDebugOverlay->isVisible();
	bool hudOverlayVisible = mHudOverlay->isVisible();

	//hide overlays
	mDebugOverlay->hide();
	mHudOverlay->hide();
	mColourFloatOverlay->hide();

	//render the depth to the RenderTexture
	mColourFloatTexture.renderTarget->update();

	//hide the scene before rendering the RenderTexture to the screen
	detachAllObjects();

	//show overlays that were visible before rendering
	if (debugOverlayVisible)
	{
		mDebugOverlay->show();
	}
	if (hudOverlayVisible)
	{
		mHudOverlay->show();
	}

	//render the result to the screen
	mColourFloatOverlay->show();
}


void Renderer::detachAllObjects()
{
	Ogre::SceneNode* rootSceneNode = mSceneManager->getRootSceneNode();

	//First walk over the children and put them all in a list.
	//Immediately detaching seems problematic, because that might break
	//the iterator, as the iterator is over attached objects and detaching
	//changes that list.
	Ogre::Node::ChildNodeIterator iter = rootSceneNode->getChildIterator();
	while (iter.hasMoreElements())
	{
		mObjectsToReattach.push_back(iter.getNext());
	}

	//now detach all the children
	size_t numObjectsToDetach = mObjectsToReattach.size();
	for (size_t i = 0; i < numObjectsToDetach; ++i)
	{
		rootSceneNode->removeChild(mObjectsToReattach[i]);
	}
}


void Renderer::reattachAllObjects()
{
	Ogre::SceneNode* rootSceneNode = mSceneManager->getRootSceneNode();

	//attach all the objects to the root
	size_t numObjectsToReattach = mObjectsToReattach.size();
	for (size_t i = 0; i < numObjectsToReattach; ++i)
	{
		rootSceneNode->addChild(mObjectsToReattach[i]);
	}

	//clear the list
	mObjectsToReattach.clear();
}